
from abc import ABC, abstractmethod
from functools import wraps
from typing import Tuple

import numpy as np
from scipy.linalg import cho_factor, solve_triangular, get_blas_funcs, block_diag
from scipy._lib._util import _asarray_validated

from autofit.graphical.utils import cached_property


class LinearOperator(ABC):
    """Implements the functionality of a linear operator. 

    All linear operators can be expressed as a tensor/matrix
    However for some it there may be more efficient representations

    e.g. The Diagonal Matrix (see `DiagonalMatrix`), 
    Sherman-Morrison Matrix (see `ShermanMorrison`)
    or the vector-Jacobian product of a function see jax.jvp [0]

    The class also has the attributes, lshape, rshape, lsize, rsize to allow multidimensional tensors to be used,
    see `ShermanMorrison`, `MultiVecOuterProduct`, or 
    `autofit.graphical.messages.transform.MultinomialLogitTransform` 
    for examples of this use case

    If `M` is the dense matrix represenation of the LinearOperator then the
    actions of the methods can be represented by the appropriate matrix
    operation, 

    Methods
    -------
    __mul__(x):
        M.dot(x) 
        M·x

    __rmul__(x):
        x.dot(M) 
        x·M

    __rtruediv__(x):
        x.dot(inv(M)) 
        x / M

    ldiv(x):
        inv(M).dot(x) 
        M \ x

    log_det():
        log(det(M))
        log |M|

    quad(x):
        x.T.dot(M).dot(x)
        xᵀ·M·x

    invquad(x):
        x.T.dot(inv(M)).dot(x)
        xᵀ·M⁻¹·x

    [0] https://jax.readthedocs.io/en/latest/notebooks/autodiff_cookbook.html#vector-jacobian-product
    """
    _ldim = 1

    @abstractmethod
    def __mul__(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def __rtruediv__(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def __rmul__(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def ldiv(self, x: np.ndarray) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def shape(self) -> Tuple[int, ...]:
        pass

    def __len__(self) -> int:
        return self.lsize

    @property
    def lsize(self) -> int:
        return np.prod(self.lshape, dtype=int)

    @property
    def rsize(self) -> int:
        return np.prod(self.rshape, dtype=int)

    @property
    def size(self) -> int:
        return np.prod(self.shape, dtype=int)

    @property
    def ldim(self):
        return self._ldim

    @property
    def rdim(self):
        return self.ndim - self.ldim

    def ndim(self):
        return len(self.shape)

    @property
    def lshape(self):
        return self.shape[:self.ldim]

    @property
    def rshape(self):
        return self.shape[self.ldim:]

    @cached_property
    @abstractmethod
    def log_det(self):
        pass

    def quad(self, M: np.ndarray) -> np.ndarray:
        return (M * self).T * self

    def invquad(self, M: np.ndarray) -> np.ndarray:
        return (M / self).T / self

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if ufunc in (np.multiply, np.matmul):
            return self.__rmul__(inputs[0])
        elif ufunc is np.divide:
            return self.__rtruediv__(inputs[0])
        else:
            return NotImplemented


class MatrixOperator(LinearOperator):

    def __init__(self, M: np.ndarray, shape=None, ldim=1):
        self.M = np.asanyarray(M)
        self._shape = shape or self.M.shape
        self._ldim = 1

    @property
    def shape(self):
        return self._shape

    @cached_property
    def log_det(self):
        return np.linalg.slogdet(self.M)[1]

    def __mul__(self, x: np.ndarray) -> np.ndarray:
        return self.M.dot(x)

    def __rtruediv__(self, x: np.ndarray) -> np.ndarray:
        return np.linalg.solve(self.M.T, x.T)

    def __rmul__(self, x: np.ndarray) -> np.ndarray:
        return np.dot(x, self.M)

    def ldiv(self, x: np.ndarray) -> np.ndarray:
        return np.linalg.solve(self.M, x)

    def to_dense(self):
        return self.M


class IdentityOperator(LinearOperator):
    def __init__(self):
        pass

    def _identity(self, values: np.ndarray) -> np.ndarray:
        return values

    __mul__ = _identity
    __rtruediv__ = _identity
    __rmul__ = _identity
    ldiv = _identity
    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    quad = _identity
    invquad = _identity

    @property
    def log_det(self):
        return 0.

    @property
    def shape(self):
        return ()

    def __len__(self):
        return 0


def _mul_triangular(c, b, trans=False, lower=True, overwrite_b=False,
                    check_finite=True):
    """wrapper for BLAS function trmv to perform triangular matrix
    multiplications


    Parameters
    ----------
    a : (M, M) array_like
        A triangular matrix
    b : (M,) or (M, N) array_like
        vector/matrix being multiplied
    lower : bool, optional
        Use only data contained in the lower triangle of `a`.
        Default is to use upper triangle.
    trans : bool, optional
        type of multiplication,

        ========  =========
        trans     system
        ========  =========
        False     a b
        True      a^T b
    overwrite_b : bool, optional
        Allow overwriting data in `b` (may enhance performance)
        not fully tested
    check_finite : bool, optional
        Whether to check that the input matrices contain only finite numbers.
        Disabling may give a performance gain, but may result in problems
        (crashes, non-termination) if the inputs do contain infinities or NaNs.
    """
    a1 = _asarray_validated(c, check_finite=check_finite)
    b1 = _asarray_validated(b, check_finite=check_finite)

    n = c.shape[1]
    if c.shape[0] != n:
        raise ValueError("Triangular matrix passed must be square")
    if b.shape[0] != n:
        raise ValueError(
            f"shapes {c.shape} and {b.shape} not aligned: "
            f"{n} (dim 1) != {b.shape[0]} (dim 0)")

    trmv, = get_blas_funcs(('trmv',), (a1, b1))
    if a1.flags.f_contiguous:
        def _trmv(a1, b1, overwrite_x):
            return trmv(
                a1, b1,
                lower=lower, trans=trans, overwrite_x=overwrite_x)
    else:
        # transposed system is solved since trmv expects Fortran ordering
        def _trmv(a1, b1, overwrite_x=overwrite_b):
            return trmv(
                a1.T, b1,
                lower=not lower, trans=not trans, overwrite_x=overwrite_x)

    if b1.ndim == 1:
        return _trmv(a1, b1, overwrite_b)
    elif b1.ndim == 2:
        # trmv only works for vector multiplications
        # set Fortran order so memory contiguous
        b2 = np.array(b1, order='F')
        for i in range(b2.shape[1]):
            # overwrite results
            _trmv(a1, b2[:, i], True)

        if overwrite_b:
            b1[:] = b2
            return b1
        else:
            return b2
    else:
        raise ValueError("b must have 1 or 2 dimensions, has {b.ndim}")


# TODO refactor these for non-square transformations

def _wrap_leftop(method):
    @wraps(method)
    def leftmethod(self, x: np.ndarray) -> np.ndarray:
        x = np.asanyarray(x)
        return method(self, x.reshape(self.rsize, -1)).reshape(x.shape)

    return leftmethod


def _wrap_rightop(method):
    @wraps(method)
    def rightmethod(self, x: np.ndarray) -> np.ndarray:
        x = np.asanyarray(x)
        return method(self, x.reshape(-1, self.lsize)).reshape(x.shape)

    return rightmethod


class CholeskyOperator(LinearOperator):
    """ This performs the whitening transforms for the passed
    cholesky factor of the Hessian/inverse covariance of the system.

    see https://en.wikipedia.org/wiki/Whitening_transformation

    >>> M = CholeskyTransform(linalg.cho_factor(hess))
    >>> y = M * x
    >>> f, df_dx = func_and_gradient(M.ldiv(y))
    >>> df_dy = df_df * M
    >>> 
    """
    # TODO implement tensor shape functionality

    def __init__(self, cho_factor):
        self.c, self.lower = self.cho_factor = cho_factor
        self.L = self.c if self.lower else self.c.T
        self.U = self.c.T if self.lower else self.c

    @classmethod
    def from_dense(cls, hess):
        return cls(cho_factor(hess))

    @_wrap_leftop
    def __mul__(self, x):
        return _mul_triangular(self.U, x, lower=False)

    @_wrap_rightop
    def __rmul__(self, x):
        return _mul_triangular(self.L, x.T, lower=True).T

    @_wrap_rightop
    def __rtruediv__(self, x):
        return solve_triangular(self.L, x.T, lower=True).T

    @_wrap_leftop
    def ldiv(self, x):
        return solve_triangular(self.U, x, lower=False)

    @cached_property
    def log_det(self):
        return np.sum(np.log(self.U.diagonal()))

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    @property
    def shape(self):
        return self.c.shape

    def to_dense(self):
        return np.tril(self.L) @ np.triu(self.U)


class InverseLinearOperator(LinearOperator):
    def __init__(self, transform):
        self.transform = transform

    @abstractmethod
    def __mul__(self, x: np.ndarray) -> np.ndarray:
        return x / self.transform

    @abstractmethod
    def __rtruediv__(self, x: np.ndarray) -> np.ndarray:
        return self.transform * x

    @abstractmethod
    def __rmul__(self, x: np.ndarray) -> np.ndarray:
        return self.transform.ldiv(x)

    @abstractmethod
    def ldiv(self, x: np.ndarray) -> np.ndarray:
        return x * self.transform

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.transform.shape

    @property
    def ldim(self):
        return self.transform.ldim

    @cached_property
    def log_det(self):
        return - self.transform.log_det

    def to_dense(self):
        return np.linalg.inv(self.transform.to_dense())


class InvCholeskyTransform(CholeskyOperator):
    """In the case where the covariance matrix is passed
    we perform the inverse operations
    """
    __mul__ = CholeskyOperator.__rtruediv__
    __rmul__ = CholeskyOperator.ldiv
    __rtruediv__ = CholeskyOperator.__mul__
    ldiv = CholeskyOperator.__rmul__

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    @cached_property
    def log_det(self):
        return - np.sum(np.log(self.U.diagonal()))

    def to_dense(self):
        return solve_triangular(self.U, np.triul(self.L), lower=False)


class DiagonalMatrix(LinearOperator):
    """
    Represents the DiagonalMatrix with diagonal `scale`

    M = np.diag(scale.ravel())
    
    """
    def __init__(self, scale, inv_scale=None):
        self.scale = np.asanyarray(scale)
        self._fscale = np.ravel(self.scale)
        self._finv_scale = 1 / \
            scale if inv_scale is None else np.ravel(inv_scale)
        self._ldim = len(self.scale.shape)

    @_wrap_leftop
    def __mul__(self, x):
        return self._fscale[:, None] * x

    @_wrap_rightop
    def __rmul__(self, x):
        return x * self._fscale

    @_wrap_rightop
    def __rtruediv__(self, x):
        return x * self._finv_scale

    @_wrap_leftop
    def ldiv(self, x):
        return self._finv_scale[:, None] * x

    @cached_property
    def log_det(self):
        return np.sum(np.log(self.scale))

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    @property
    def shape(self):
        return self.scale.shape * 2

    def to_dense(self):
        return np.diag(self._fscale).reshape(self.shape)


class VecOuterProduct(LinearOperator):
    """
    represents the matrix vector outer product

    outer = vec[:, None] * vecT[None, :]
    """
    def __init__(self, vec, vecT=None):
        self.vec = np.asanyarray(vec)
        self.vecT = np.asanyarray(vec if vecT is None else vecT)
        self._fvec = np.ravel(self.vec)[:, None]
        self._fvecT = np.ravel(self.vecT)[None, :]

    @_wrap_leftop
    def __mul__(self, x):
        return self._fvec @ self._fvecT.dot(x)

    @_wrap_rightop
    def __rmul__(self, x):
        return x.dot(self._fvec) @ self._fvecT

    def __rtruediv__(self, x):
        raise NotImplementedError()

    def ldiv(self, x):
        raise NotImplementedError()

    @cached_property
    def log_det(self):
        return - np.inf

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    __rmatmul__ = __rmul__

    @property
    def shape(self):
        return (self.vec.size, self.vecT.size)

    @property
    def ldim(self):
        return len(self.vec.shape)

    def to_dense(self):
        return np.outer(self._fvec, self._fvecT).reshape(self.shape)


class MultiVecOuterProduct(LinearOperator):
    """
    represents the matrix vector outer product for stacked vectors,

    outer -> block_diag(*
        (v[:, None] * u[None, :] for v, u in zip(vec, vecT)
    )
    outer @ x -> np.vstack([
        v[:, None] * u[None, :] @ x for v, u in zip(vec, vecT)
    ])
    """

    def __init__(self, vec, vecT=None):
        self.vec = np.asanyarray(vec)
        self.vecT = np.asanyarray(vec if vecT is None else vecT)
        self.n, self.d = self.vec.shape
        self.nT, self.dT = self.vecT.shape

    @_wrap_leftop
    def __mul__(self, x):
        return np.einsum(
            "ij,ik,ikl -> ijl",
            self.vec, self.vecT, x.reshape(*self.lshape, -1)
        )

    @_wrap_rightop
    def __rmul__(self, x):
        return np.einsum(
            "ij,ik,lij -> lik",
            self.vec, self.vecT, x.reshape(-1, *self.rshape)
        )

    def __rtruediv__(self, x):
        raise NotImplementedError()

    def ldiv(self, x):
        raise NotImplementedError()

    @cached_property
    def log_det(self):
        return - np.inf

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    __rmatmul__ = __rmul__

    @property
    def shape(self):
        return self.vec.shape + self.vecT.shape

    @property
    def lshape(self):
        return self.shape[:2]

    @property
    def rshape(self):
        return self.shape[2:]

    def to_dense(self):
        return block_diag(*(
            self.vec[:, :, None] * self.vecT[:, None, :]
        )).reshape(self.shape)


class ShermanMorrison(LinearOperator):
    """
    Represents the Sherman-Morrison low rank update, 
    inv(A + vec @ vec.T) = 
        inv(A) - inv(A) @ vec @ vec.T @ inv(A) / (1 + vec @ inv(A) @ vec/T)
    """

    def __init__(self, linear, vec):
        self.linear = linear
        if np.ndim(vec) == 2:
            self.outer = MultiVecOuterProduct(vec)
        elif np.ndim(vec) == 1:
            self.outer = VecOuterProduct(vec)
        else:
            raise ValueError("vec must be 1 or 2 dimensional")

        self.inv_scale = 1 + linear.quad(vec)
        self._ldim = self.linear.ldim

    @_wrap_leftop
    def __mul__(self, x):
        return self.linear * x + self.outer * x

    @_wrap_rightop
    def __rmul__(self, x):
        return x * self.linear + x * self.outer

    @_wrap_rightop
    def __rtruediv__(self, x):
        x1 = x / self.linear
        return x1 - ((x1/self.inv_scale) * self.outer) / self.linear

    @_wrap_leftop
    def ldiv(self, x):
        x1 = self.linear.ldiv(x)
        return x1 - self.linear.ldiv(self.outer.dot(x1/self.inv_scale))

    @cached_property
    def log_det(self):
        return self.linear.log_det + np.log(self.inv_scale)

    @property
    def shape(self):
        return self.linear.shape

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    __rmatmul__ = __rmul__

    def to_dense(self):
        dense_outer = self.outer.to_dense()
        return self.linear.to_dense().reshape(self.shape) + dense_outer



import pickle

import autofit as af


def test_serialize_optimiser():
    optimizer = af.DynestyStatic(
        "name"
    )

    optimizer = pickle.loads(
        pickle.dumps(
            optimizer
        )
    )
    assert optimizer.name == "name"



from typing import Optional

import numpy as np
from dynesty.results import Results

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples import Sample
from autofit import NestSamples


class DynestySamples(NestSamples):

    def __init__(
            self,
            model: AbstractPriorModel,
            results: Results,
            number_live_points: int,
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
    ):
        """
        The *Output* classes in **PyAutoFit** provide an interface between the results of a `NonLinearSearch` (e.g.
        as files on your hard-disk) and Python.

        For example, the output class can be used to load an instance of the best-fit model, get an instance of any
        individual sample by the `NonLinearSearch` and return information on the likelihoods, errors, etc.

        The Bayesian log evidence estimated by the nested sampling algorithm.

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        """

        self.results = results
        self._number_live_points = number_live_points

        parameter_lists = self.results.samples.tolist()
        log_prior_list = [
            sum(model.log_prior_list_from_vector(vector=vector)) for vector in parameter_lists
        ]
        log_likelihood_list = list(self.results.logl)

        try:
            weight_list = list(
                np.exp(np.asarray(self.results.logwt) - self.results.logz[-1])
            )
        except:
            weight_list = self.results["weights"]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list
        )

        super().__init__(
            model=model,
            sample_list=sample_list,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
        )

    @property
    def number_live_points(self):
        return self._number_live_points

    @property
    def total_samples(self):
        return int(np.sum(self.results.ncall))

    @property
    def log_evidence(self):
        return np.max(self.results.logz)



import math
from typing import List, Optional

import numpy as np

from autofit.mapper.model_mapper import ModelMapper
from autofit.non_linear.mcmc.auto_correlations import AutoCorrelationsSettings
from autofit.non_linear.samples.pdf import PDFSamples
from .optimizer import OptimizerSamples
from .sample import Sample, load_from_table


class MCMCSamples(PDFSamples):

    def __init__(
            self,
            model: ModelMapper,
            sample_list: List[Sample],
            auto_correlation_settings: AutoCorrelationsSettings,
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
    ):

        self.auto_correlation_settings = auto_correlation_settings

        super().__init__(
            model=model,
            sample_list=sample_list,
            unconverged_sample_size=unconverged_sample_size,
            time=time,
        )

    @property
    def total_walkers(self):
        raise NotImplementedError

    @property
    def total_steps(self):
        raise NotImplementedError

    @property
    def auto_correlations(self):
        raise NotImplementedError

    @classmethod
    def from_table(self, filename: str, model: ModelMapper, number_live_points: int = None):
        """
        Write a table of parameters, posteriors, priors and likelihoods

        Parameters
        ----------
        filename
            Where the table is to be written
        """

        sample_list = load_from_table(filename=filename)

        return OptimizerSamples(
            model=model,
            sample_list=sample_list
        )

    @property
    def info_json(self):
        return {
            "times": None,
            "check_size": self.auto_correlations.check_size,
            "required_length": self.auto_correlations.required_length,
            "change_threshold": self.auto_correlations.change_threshold,
            "total_walkers": self.total_walkers,
            "total_steps": self.total_steps,
            "time": self.time,
        }

    @property
    def pdf_converged(self):
        """
        To analyse and visualize samples using corner.py, the analysis must be sufficiently converged to produce
        smooth enough PDF for analysis. This property checks whether the non-linear search's samples are sufficiently
        converged for corner.py use.

        Emcee samples can be analysed by corner.py irrespective of how long the sampler has run, albeit low run times
        will likely produce inaccurate results.
        """
        try:
            samples_after_burn_in = self.samples_after_burn_in
            if len(samples_after_burn_in) == 0:
                return False
            return True
        except ValueError:
            return False

    @property
    def samples_after_burn_in(self) -> [list]:
        """
        The emcee samples with the initial burn-in samples removed.

        The burn-in period is estimated using the auto-correlation times of the parameters.
        """
        raise NotImplementedError()

    @property
    def converged(self) -> bool:
        """
        Whether the emcee samples have converged on a solution or if they are still in a burn-in period, based on the
        auto correlation times of parameters.
        """
        return self.auto_correlations.check_if_converged(
            total_samples=self.total_samples
        )

    @property
    def median_pdf_vector(self) -> [float]:
        """
        The median of the probability density function (PDF) of every parameter marginalized in 1D, returned
        as a list of values.

        This is computed by binning all sampls after burn-in into a histogram and take its median (e.g. 50%) value.
        """

        if self.pdf_converged:
            return [
                float(np.percentile(self.samples_after_burn_in[:, i], [50]))
                for i in range(self.model.prior_count)
            ]

        return self.max_log_likelihood_vector

    def vector_at_sigma(self, sigma: float) -> [float]:
        """
        The value of every parameter marginalized in 1D at an input sigma value of its probability density function
        (PDF), returned as two lists of values corresponding to the lower and upper values parameter values.

        For example, if sigma is 1.0, the marginalized values of every parameter at 31.7% and 68.2% percentiles of each
        PDF is returned.

        This does not account for covariance between parameters. For example, if two parameters (x, y) are degenerate
        whereby x decreases as y gets larger to give the same PDF, this function will still return both at their
        upper values. Thus, caution is advised when using the function to reperform a model-fits.

        For Emcee, if the samples have converged this is estimated by binning the samples after burn-in into a
        histogram and taking the parameter values at the input PDF %.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        limit = math.erf(0.5 * sigma * math.sqrt(2))

        if self.pdf_converged:
            samples = self.samples_after_burn_in

            return [
                tuple(
                    np.percentile(samples[:, i], [100.0 * (1.0 - limit), 100.0 * limit])
                )
                for i in range(self.model.prior_count)
            ]

        parameters_min = list(
            np.min(self.parameter_lists[-self.unconverged_sample_size:], axis=0)
        )
        parameters_max = list(
            np.max(self.parameter_lists[-self.unconverged_sample_size:], axis=0)
        )

        return [
            (parameters_min[index], parameters_max[index])
            for index in range(len(parameters_min))
        ]

    @property
    def log_evidence(self):
        return None



from uuid import uuid4

import autofit as af
from autofit.mock.mock import Gaussian


def test_float_inequality(session):
    aggregator = af.Aggregator(session)

    for sigma in [
        0.9992237362814176,
        4.9687212446221904,
        9.967065800134504,
    ]:
        session.add(
            af.db.Fit(
                id=str(uuid4()),
                instance={
                    "gaussian": Gaussian(
                        sigma=sigma
                    )
                }
            )
        )
        session.commit()

    assert len(aggregator) == 3

    assert len(aggregator.query(
        aggregator.model.gaussian.sigma < 3
    )) == 1



import numpy as np
from contextlib import contextmanager
import sys, os
import json

import builtins
import importlib
import re
from typing import List, Type


def get_class_path(cls: type) -> str:
    """
    The full import path of the type
    """
    return re.search("'(.*)'", str(cls))[1]


def get_class(class_path: str) -> Type[object]:
    return GetClass(
        class_path
    ).cls


class GetClass:
    def __init__(self, class_path):
        self.class_path = class_path

    @property
    def _class_path_array(self) -> List[str]:
        """
        A list of strings describing the module and class of the
        real object represented here
        """
        return self.class_path.split(".")

    @property
    def _class_name(self) -> str:
        """
        The name of the real class
        """
        return self._class_path_array[-1]

    @property
    def _module_path(self) -> str:
        """
        The path of the module containing the real class
        """
        return ".".join(self._class_path_array[:-1])

    @property
    def _module(self):
        """
        The module containing the real class
        """
        try:
            return importlib.import_module(
                self._module_path
            )
        except ValueError:
            return builtins

    @property
    def cls(self) -> Type[object]:
        """
        The class of the real object
        """
        return getattr(
            self._module,
            self._class_name
        )

@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout

def numpy_array_to_json(
    array: np.ndarray, file_path: str, overwrite: bool = False
):
    """
    Write a NumPy array to a json file.

    Parameters
    ----------
    array : np.ndarray
        The array that is written to json.
    file_path : str
        The full path of the file that is output, including the file name and `.json` extension.
    overwrite : bool
        If `True` and a file already exists with the input file_path the .json file is overwritten. If 
        `False`, an error will be raised.

    Returns
    -------
    None

    Examples
    --------
    array_2d = np.ones((5,5))
    numpy_array_to_json(array_2d=array_2d, file_path='/path/to/file/filename.json', overwrite=True)
    """

    file_dir = os.path.split(file_path)[0]

    if not os.path.exists(file_dir):
        os.makedirs(file_dir)

    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    with open(file_path, "w+") as f:
        json.dump(array.tolist(), f)

def numpy_array_from_json(file_path: str):
    """
    Read a 1D NumPy array from a .json file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the structures
    appear the same orientation as .json files loaded in DS9.

    Parameters
    ----------
    file_path : str
        The full path of the file that is loaded, including the file name and ``.json`` extension.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .json file.

    Examples
    --------
    array_2d = numpy_array_from_json(file_path='/path/to/file/filename.json')
    """
    with open(file_path, "r") as f:
        return np.asarray(json.load(f))



import os
import shutil
from os import path

import pytest
from matplotlib import pyplot
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import autofit as af
import autofit.non_linear.samples.sample
import autofit.non_linear.samples.stored
from autoconf import conf
from autofit import database as db
from autofit.mock import mock

directory = path.dirname(path.realpath(__file__))


class PlotPatch:
    def __init__(self):
        self.paths = []

    def __call__(self, path, *args, **kwargs):
        self.paths.append(path)


@pytest.fixture(name="plot_patch")
def make_plot_patch(monkeypatch):
    plot_patch = PlotPatch()
    monkeypatch.setattr(pyplot, "savefig", plot_patch)
    return plot_patch


@pytest.fixture(name="session")
def make_session():
    engine = create_engine('sqlite://')
    session = sessionmaker(bind=engine)()
    db.Base.metadata.create_all(engine)
    yield session
    session.close()
    engine.dispose()


@pytest.fixture(
    autouse=True,
    scope="session"
)
def remove_logs():
    yield
    for d, _, files in os.walk(directory):
        for file in files:
            if file.endswith(".log"):
                os.remove(path.join(d, file))


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance.push(
        new_path=path.join(directory, "config"),
        output_path=path.join(directory, "output")
    )


@pytest.fixture
def model():
    return af.ModelMapper()


@pytest.fixture(name="samples")
def make_samples():
    sample_0 = autofit.non_linear.samples.sample.Sample(log_likelihood=1.0, log_prior=2.0, weight=0.25)
    sample_1 = autofit.non_linear.samples.sample.Sample(log_likelihood=3.0, log_prior=5.0, weight=0.75)

    return autofit.non_linear.samples.stored.StoredSamples(
        model=af.Mapper(),
        samples=[sample_0, sample_1],
    )


@pytest.fixture(name="result")
def make_result():
    model = af.Mapper()
    model.one = af.PriorModel(mock.MockComponents, component=mock.MockClassx2)
    return af.Result(model=model, samples=mock.MockSamples(), search=mock.MockSearch())


@pytest.fixture(name="collection")
def make_collection():
    collection = af.ResultsCollection()
    model = af.ModelMapper()
    model.one = af.PriorModel(mock.MockComponents, component=mock.MockClassx2)
    instance = af.ModelInstance()
    instance.one = mock.MockComponents(component=mock.MockClassx2())

    result = af.MockResult(model=model, instance=instance)

    model = af.ModelMapper()
    instance = af.ModelInstance()

    model.hyper_galaxy = mock.HyperGalaxy
    instance.hyper_galaxy = mock.HyperGalaxy()

    hyper_result = af.MockResult(model=model, instance=instance)

    result.hyper_result = hyper_result

    collection.add("search name", result)

    return collection



from pathlib import Path

import pytest

from autofit.tools.edenise import Package

package_directory = Path(
    __file__
).parent.parent.parent.parent / "autofit"


@pytest.fixture(
    name="package"
)
def make_package():
    return Package(
        package_directory,
        prefix="VIS_CTI",
        is_top_level=True,
        eden_dependencies=["autoconf"],
        should_rename_modules=True
    )



from os import path

import pytest

import autofit as af
from autofit.mock.mock import MockClassx4
from autofit.non_linear.samples import OptimizerSamples, Sample

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(name="samples")
def make_samples():
    model = af.ModelMapper(mock_class_1=MockClassx4)

    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]

    return OptimizerSamples(
        model=model,
        samples=Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[1.0, 2.0, 3.0, 10.0, 5.0],
            log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
            weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
        )
    )




from .wrap_base import (
    Units,
    Figure,
    Axis,
    Cmap,
    Colorbar,
    ColorbarTickParams,
    TickParams,
    YTicks,
    XTicks,
    Title,
    YLabel,
    XLabel,
    Legend,
    Output,
)



import pytest

import autofit as af
import autofit.graphical as g
from autoconf.conf import output_path_for_test
from autofit.graphical.utils import Status
from test_autofit.graphical.gaussian.model import Analysis


@pytest.fixture(
    name="analysis"
)
def make_analysis(
        x, y
):
    return Analysis(
        x=x,
        y=y
    )


@pytest.fixture(
    name="factor_model"
)
def make_factor_model(
        prior_model,
        analysis
):
    return g.AnalysisFactor(
        prior_model,
        analysis=analysis
    )


@pytest.fixture(
    name="laplace"
)
def make_laplace():
    return g.LaplaceFactorOptimiser()


@pytest.fixture(
    name="dynesty"
)
def make_dynesty():
    return af.DynestyStatic(
        maxcall=10
    )


def test_default(
        factor_model,
        laplace
):
    model = factor_model.optimise(laplace)

    assert model.centre.mean == pytest.approx(50, rel=0.1)
    assert model.intensity.mean == pytest.approx(25, rel=0.1)
    assert model.sigma.mean == pytest.approx(10, rel=0.1)


def test_set_model_identifier(
        dynesty,
        prior_model,
        analysis
):
    dynesty.fit(
        prior_model,
        analysis
    )

    identifier = dynesty.paths.identifier
    assert identifier is not None

    prior_model.centre = af.GaussianPrior(mean=20, sigma=20)
    dynesty.fit(
        prior_model,
        analysis
    )

    assert identifier != dynesty.paths.identifier


class TestDynesty:
    @output_path_for_test()
    def test_optimisation(
            self,
            factor_model,
            laplace,
            dynesty
    ):
        factor_model.optimiser = dynesty
        factor_model.optimise(laplace)

    @output_path_for_test()
    def test_optimise(
            self,
            factor_model,
            dynesty
    ):
        result, status = dynesty.optimise(
            factor_model,
            factor_model.mean_field_approximation()
        )

        assert isinstance(
            result,
            g.EPMeanField
        )
        assert isinstance(
            status,
            Status
        )



import pytest

from autofit.tools.edenise import LineItem


@pytest.fixture(
    autouse=True
)
def set_flag(package):
    package._should_remove_type_annotations = True


def test_dotted_annotation(
        package
):
    assert LineItem(
        "def my_func() -> np.ndarray:",
        parent=package
    ).target_string == "def my_func():"


def test_ellipsis(
        package
):
    assert LineItem(
        "def my_func(t: Tuple[int, ...]) -> Tuple[int, ...]:",
        parent=package
    ).target_string == "def my_func(t):"


def test_complex_example(
        package
):
    assert LineItem(
        """def grid_2d_radial_projected_from(
    self, centre: Tuple[float, float] = (0.0, 0.0), angle: float = 0.0
) -> grid_2d_irregular.Grid2DIrregular:""",
        parent=package
    ).target_string == """def grid_2d_radial_projected_from(
    self, centre= (0.0, 0.0), angle= 0.0
):"""


def test_lost_argument(
        package
):
    assert LineItem(
        "def convert_grid_2d(grid_2d: Union[np.ndarray, List], mask_2d):",
        parent=package
    ).target_string == "def convert_grid_2d(grid_2d, mask_2d):"


def test_is_function(
        package
):
    assert LineItem(
        "def my_func():",
        parent=package
    ).is_function is True


def test_regression(
        package
):
    assert LineItem(
        """def path_instances_of_class(
        obj, cls: type, ignore_class: Optional[Union[type, Tuple[type]]] = None
):""",
        parent=package
    ).target_string == """def path_instances_of_class(
        obj, cls, ignore_class= None
):"""


class TestStripAnnotations:
    @pytest.mark.parametrize(
        "string",
        [
            "def my_func() -> dict:",
            "def my_func()->dict:",
            "def my_func() -> dict :",
        ]
    )
    def test_strip_return_type(
            self,
            package,
            string
    ):
        assert LineItem(
            string,
            parent=package
        ).target_string == "def my_func():"

    def test_across_new_lines(
            self,
            package
    ):
        line_item = LineItem(
            """def my_func(
                one: dict,
                two: dict
            ):
            """,
            parent=package
        )
        assert line_item.is_function
        assert line_item.target_string == """def my_func(
                one,
                two
            ):
            """

    @pytest.mark.parametrize(
        "annotation",
        [
            "Optional[Union[list, str]]",
            "Optional[Union[type, Tuple[type]]]"
        ]
    )
    def test_complex_type_annotation(
            self,
            package,
            annotation
    ):
        assert LineItem(
            f"def my_func(complex: {annotation}):",
            parent=package
        ).target_string == "def my_func(complex):"

    def test_dont_convert_dict(
            self,
            package
    ):
        string = "{'one': 1, 'two': 2}"
        assert LineItem(
            string,
            parent=package
        ).target_string == string

    @pytest.mark.parametrize(
        "string",
        [
            "def my_func() -> dict:",
            "def my_func()->dict:",
            "def my_func() -> dict :",
        ]
    )
    def test_strip_return_type(
            self,
            package,
            string
    ):
        assert LineItem(
            string,
            parent=package
        ).target_string == "def my_func():"

    def test_multiple_arguments(
            self,
            package
    ):
        assert LineItem(
            "def my_func(arg1: dict, arg2: dict):",
            parent=package
        ).target_string == "def my_func(arg1, arg2):"

    def test_strip_argument_type(
            self,
            package
    ):
        assert LineItem(
            "def my_func(arg: dict):",
            parent=package
        ).target_string == "def my_func(arg):"



from abc import ABC, abstractmethod
from functools import reduce, wraps
from inspect import getfullargspec
from itertools import count
from numbers import Real
from operator import and_
from typing import Optional, Tuple, Union, Iterator, Type, List

import numpy as np

from autofit.mapper.prior.prior import Prior, GaussianPrior
from .transform import AbstractDensityTransform, LinearShiftTransform
from ..factor_graphs.jacobians import FactorJacobian
from ..utils import cached_property
from ...mapper.variable import Variable


def assert_ids_match(func):
    @wraps(func)
    def wrapper(self, other):
        if isinstance(
                other,
                AbstractMessage
        ) and self.id != other.id:
            raise AssertionError(
                f"Message with id {self.id} should not be compared to message with id {other.id}"
            )
        result = func(self, other)
        return result

    return wrapper


class AbstractMessage(ABC):
    log_base_measure: float
    _projection_class: Optional[Type['AbstractMessage']] = None
    _multivariate: bool = False
    _parameter_support: Optional[Tuple[Tuple[float, float], ...]] = None
    _support: Optional[Tuple[Tuple[float, float], ...]] = None

    _ids = count()

    def __init__(
            self,
            *parameters: Union[np.ndarray, float],
            log_norm=0.,
            id_=None
    ):
        self.id_ = id_ or next(self._ids)
        self.log_norm = log_norm
        self._broadcast = np.broadcast(*parameters)
        self.id = id_
        if self.shape:
            self.parameters = tuple(
                np.asanyarray(p) for p in parameters)
        else:
            self.parameters = tuple(parameters)

    @classmethod
    def from_prior(cls, prior: Prior) -> "AbstractMessage":
        from autofit.graphical import NormalMessage
        if isinstance(
                prior, GaussianPrior
        ):
            return NormalMessage(
                mu=prior.mean,
                sigma=prior.sigma,
                id_=prior.id
            )
        raise TypeError(
            f"No message exists for prior of type {type(prior)}"
        )

    @cached_property
    @abstractmethod
    def natural_parameters(self):
        pass

    @abstractmethod
    def sample(self, n_samples: Optional[int] = None):
        pass

    @staticmethod
    @abstractmethod
    def invert_natural_parameters(natural_parameters: np.ndarray
                                  ) -> Tuple[np.ndarray, ...]:
        pass

    @staticmethod
    @abstractmethod
    def to_canonical_form(x: np.ndarray) -> np.ndarray:
        pass

    @cached_property
    @abstractmethod
    def log_partition(self) -> np.ndarray:
        pass

    @cached_property
    @abstractmethod
    def mean(self) -> np.ndarray:
        pass

    @cached_property
    @abstractmethod
    def variance(self) -> np.ndarray:
        pass

    @cached_property
    def scale(self) -> np.ndarray:
        return self.variance ** 0.5

    def __hash__(self):
        return self.id or super().__hash__()

    @classmethod
    def calc_log_base_measure(cls, x):
        return cls.log_base_measure

    def __iter__(self) -> Iterator[np.ndarray]:
        return iter(self.parameters)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._broadcast.shape

    @property
    def size(self) -> int:
        return self._broadcast.size

    @property
    def ndim(self) -> int:
        return self._broadcast.ndim

    @classmethod
    def from_natural_parameters(
            cls,
            natural_parameters: np.ndarray,
            **kwargs
    ) -> "AbstractMessage":
        cls = cls._projection_class or cls
        args = cls.invert_natural_parameters(natural_parameters)
        return cls(*args, **kwargs)

    @classmethod
    @abstractmethod
    def invert_sufficient_statistics(
            cls, sufficient_statistics: np.ndarray
    ) -> np.ndarray:
        pass

    @classmethod
    def from_sufficient_statistics(
            cls, suff_stats: np.ndarray, **kwargs
    ) -> "AbstractMessage":
        natural_params = cls.invert_sufficient_statistics(suff_stats)
        cls = cls._projection_class or cls
        return cls.from_natural_parameters(natural_params, **kwargs)

    def sum_natural_parameters(
            self, *dists: "AbstractMessage"
    ) -> "AbstractMessage":
        """return the unnormalised result of multiplying the pdf
        of this distribution with another distribution of the same
        type
        """
        new_params = sum(
            (dist.natural_parameters for dist in self._iter_dists(dists)
             if isinstance(dist, AbstractMessage)),
            self.natural_parameters)
        mul_dist = self.from_natural_parameters(
            new_params,
            id_=self.id
        )
        return mul_dist

    @assert_ids_match
    def sub_natural_parameters(
            self, other: "AbstractMessage"
    ) -> "AbstractMessage":
        """return the unnormalised result of dividing the pdf
        of this distribution with another distribution of the same
        type"""
        log_norm = self.log_norm - other.log_norm
        new_params = (self.natural_parameters - other.natural_parameters)
        div_dist = self.from_natural_parameters(
            new_params,
            log_norm=log_norm,
            id_=self.id
        )
        return div_dist

    _multiply = sum_natural_parameters
    _divide = sub_natural_parameters

    @assert_ids_match
    def __mul__(self, other: Union["AbstractMessage", Real]) -> "AbstractMessage":
        if isinstance(other, AbstractMessage):
            return self._multiply(other)
        else:
            log_norm = self.log_norm + np.log(other)
            return type(self)(
                *self.parameters,
                log_norm=log_norm,
                id_=self.id
            )

    @assert_ids_match
    def __rmul__(self, other: "AbstractMessage") -> "AbstractMessage":
        return self * other

    @assert_ids_match
    def __truediv__(self, other: Union["AbstractMessage", Real]) -> "AbstractMessage":
        if isinstance(other, AbstractMessage):
            return self._divide(other)
        else:
            log_norm = self.log_norm - np.log(other)
            return type(self)(
                *self.parameters,
                log_norm=log_norm,
                id_=self.id
            )

    @assert_ids_match
    def __pow__(self, other: Real) -> "AbstractMessage":
        natural = self.natural_parameters
        new_params = other * natural
        log_norm = other * self.log_norm
        return self.from_natural_parameters(
            new_params,
            log_norm=log_norm,
            id_=self.id
        )

    @classmethod
    def parameter_names(cls):
        return getfullargspec(cls.__init__).args[1:-1]

    def __str__(self) -> str:
        param_attrs = [
            (attr, np.asanyarray(getattr(self, attr)))
            for attr in self.parameter_names()]
        if self.shape:
            pad = max(len(attr) for attr, _ in param_attrs)
            attr_str = "    {:<%d}={}" % pad
            param_strs = ',\n'.join(
                attr_str.format(
                    attr, np.array2string(val, prefix=' ' * (pad + 5)))
                for attr, val in param_attrs)
            return f"{type(self).__name__}(\n{param_strs})"
        else:
            param_strs = ', '.join(
                attr + '=' +
                np.array2string(val, prefix=' ' * (len(attr) + 1))
                for attr, val in param_attrs)
            return f"{type(self).__name__}({param_strs})"

    __repr__ = __str__

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return np.exp(self.logpdf(x))

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        shape = np.shape(x)
        if shape:
            x = np.asanyarray(x)

        if shape == self.shape:
            eta = self.natural_parameters
            t = self.to_canonical_form(x)
            log_base = self.calc_log_base_measure(x)
            # TODO this can be made more efficient using tensordot
            eta_t = np.multiply(eta, t).sum(0)
            return log_base + eta_t - self.log_partition

        elif shape[1:] == self.shape:
            eta = self.natural_parameters
            t = self.to_canonical_form(x)
            log_base = self.calc_log_base_measure(x)
            eta_t = np.multiply(eta[:, None, ...], t).sum(0)
            return log_base + eta_t - self.log_partition

        raise ValueError(
            f"shape of passed value {shape} does not "
            f"match message shape {self.shape}")

    def numerical_logpdf_gradient(self, x: np.ndarray, eps: float = 1e-6
                                  ) -> Tuple[np.ndarray, np.ndarray]:
        shape = np.shape(x)
        if shape:
            x0 = np.array(x, dtype=np.float64)
            logl0 = self.logpdf(x0)
            if self._multivariate:
                grad_logl = np.empty(logl0.shape + x0.shape)
                sl = tuple(slice(None) for _ in range(logl0.ndim))
                with np.nditer(x0, flags=['multi_index'], op_flags=['readwrite']) as it:
                    for xv in it:
                        xv += eps
                        logl = self.logpdf(x0)
                        grad_logl[sl + it.multi_index] = (logl - logl0) / eps
                        xv -= eps
            else:
                l0 = logl0.sum()
                grad_logl = np.empty_like(x0)
                with np.nditer(x0, flags=['multi_index'], op_flags=['readwrite']) as it:
                    for xv in it:
                        xv += eps
                        logl = self.logpdf(x0).sum()  # type: ignore
                        grad_logl[it.multi_index] = (logl - l0) / eps
                        xv -= eps
        else:
            logl0 = self.logpdf(x)
            grad_logl = (self.logpdf(x + eps) - logl0) / eps

        return logl0, grad_logl

    def numerical_logpdf_gradient_hessian(self, x: np.ndarray, eps: float = 1e-6
                                          ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        shape = np.shape(x)
        if shape:
            x0 = np.array(x, dtype=np.float64)
            if self._multivariate:
                logl0, gradl0 = self.numerical_logpdf_gradient(x0)
                hess_logl = np.empty(gradl0.shape + x0.shape)
                sl = tuple(slice(None) for _ in range(gradl0.ndim))
                with np.nditer(x0, flags=['multi_index'], op_flags=['readwrite']) as it:
                    for xv in it:
                        xv += eps
                        _, gradl = self.numerical_logpdf_gradient(x0)
                        hess_logl[sl + it.multi_index] = (gradl - gradl0) / eps
                        xv -= eps
            else:
                logl0 = self.logpdf(x0)
                l0 = logl0.sum()
                grad_logl = np.empty_like(x0)
                hess_logl = np.empty_like(x0)
                with np.nditer(x0, flags=['multi_index'], op_flags=['readwrite']) as it:
                    for xv in it:
                        xv += eps
                        l1 = self.logpdf(x0).sum()
                        xv -= 2 * eps
                        l2 = self.logpdf(x0).sum()
                        g1 = (l1 - l0) / eps
                        g2 = (l0 - l2) / eps
                        grad_logl[it.multi_index] = g1
                        hess_logl[it.multi_index] = (g1 - g2) / eps
                        xv += eps

                gradl0 = grad_logl
        else:
            logl0 = self.logpdf(x)
            logl1 = self.logpdf(x + eps)
            logl2 = self.logpdf(x - eps)
            gradl0 = (logl1 - logl0) / eps
            gradl1 = (logl0 - logl2) / eps
            hess_logl = (gradl0 - gradl1) / eps

        return logl0, gradl0, hess_logl

    logpdf_gradient = numerical_logpdf_gradient
    logpdf_gradient_hessian = numerical_logpdf_gradient_hessian

    @classmethod
    def project(
            cls,
            samples: np.ndarray,
            log_weight_list: Optional[np.ndarray] = None,
            id_=None
    ) -> "AbstractMessage":
        """Calculates the sufficient statistics of a set of samples
        and returns the distribution with the appropriate parameters
        that match the sufficient statistics
        """
        # if weight_list aren't passed then equally weight all samples

        # Numerically stable weighting for very small/large weight_list

        # rescale coordinates to 'natural parameter space'
        if log_weight_list is None:
            log_weight_list = np.zeros_like(samples)

        log_w_max = np.max(log_weight_list, axis=0, keepdims=True)
        w = np.exp(log_weight_list - log_w_max)
        norm = w.mean(0)
        log_norm = np.log(norm) + log_w_max[0]

        tx = cls.to_canonical_form(samples)
        w /= norm
        suff_stats = (tx * w[None, ...]).mean(1)

        assert np.isfinite(suff_stats).all()

        cls = cls._projection_class or cls
        return cls.from_sufficient_statistics(
            suff_stats,
            log_norm=log_norm,
            id_=id_
        )

    def has_exact_projection(self, x: "AbstractMessage") -> bool:
        return type(self) == type(x)

    def calc_exact_projection(self, x: "AbstractMessage"):
        if type(self) == type(x):
            projection = self * x
            projection.log_norm = self.log_normalisation(x)
            return {'x': projection}
        raise NotImplementedError()

    def calc_exact_update(self, x: "AbstractMessage"):
        if type(self) == type(x):
            log_norm = self.log_normalisation(x)
            return {'x': type(x)(*self.parameters, log_norm=log_norm)}
        else:
            raise NotImplementedError()

    @classmethod
    def from_mode(
            cls,
            mode: np.ndarray,
            covariance: np.ndarray,
            id_
    ) -> "AbstractMessage":
        pass

    def log_normalisation(self, *elems: Union["AbstractMessage", float]
                          ) -> np.ndarray:
        """
        Calculates the log of the integral of the product of a
        set of distributions

        NOTE: ignores log normalisation
        """
        # Remove floats from messages passed
        dists: List[AbstractMessage] = [
            dist for dist in self._iter_dists(elems)
            if isinstance(dist, AbstractMessage)]

        # Calculate log product of message normalisation
        log_norm = self.log_base_measure - self.log_partition
        log_norm += sum(
            dist.log_base_measure - dist.log_partition for dist in dists)

        # Calculate log normalisation of product of messages
        prod_dist = self.sum_natural_parameters(*dists)
        log_norm -= prod_dist.log_base_measure - prod_dist.log_partition

        return log_norm

    @staticmethod
    def _iter_dists(dists) -> Iterator[Union["AbstractMessage", float]]:
        for elem in dists:
            if isinstance(elem, AbstractMessage):
                yield elem
            elif np.isscalar(elem):
                yield elem
            else:
                for dist in elem:
                    yield dist

    def update_invalid(self, other: 'AbstractMessage') -> 'AbstractMessage':
        valid = self.check_valid()
        if self.ndim:
            valid_parameters: Iterator[np.ndarray] = (
                np.where(valid, p, p_safe) for p, p_safe in zip(self, other))
        else:
            # TODO: Fairly certain this would not work
            valid_parameters = iter(self if valid else other)
        return type(self)(
            *valid_parameters,
            log_norm=self.log_norm,
            id_=self.id
        )

    def check_support(self) -> np.ndarray:
        if self._parameter_support is not None:
            return reduce(
                and_,
                ((p >= support[0]) & (p <= support[1])
                 for p, support in
                 zip(self.parameters, self._parameter_support)))
        elif self.ndim:
            return np.array(True, dtype=bool, ndmin=self.ndim)
        return np.array([True])

    def check_finite(self) -> np.ndarray:
        return np.isfinite(self.natural_parameters).all(0)

    def check_valid(self) -> np.ndarray:
        return self.check_finite() & self.check_support()

    @cached_property
    def is_valid(self) -> np.bool_:
        return np.all(self.check_finite()) and np.all(self.check_support())

    @staticmethod
    def _get_mean_variance(mean: np.ndarray, covariance: np.ndarray
                           ) -> Tuple[np.ndarray, np.ndarray]:
        mean, covariance = np.asanyarray(mean), np.asanyarray(covariance)

        if not covariance.shape:
            # If variance is float simply pass through
            variance = covariance * np.ones_like(mean)
            if not variance.shape:
                variance = variance.item()
        elif mean.shape == covariance.shape:
            variance = np.asanyarray(covariance)
        elif covariance.shape == mean.shape * 2:
            # If 2D covariance matrix passed get diagonal
            inds = tuple(np.indices(mean.shape))
            variance = np.asanyarray(covariance)[inds * 2]
        else:
            raise ValueError(
                f"shape of covariance {covariance.shape} is invalid "
                f"must be (), {mean.shape}, or {mean.shape * 2}")
        return mean, variance

    def __call__(
            self,
            x: np.ndarray,
            _variables: Optional[Tuple[str]] = ('x',)
    ) -> Union[np.ndarray, Tuple[np.ndarray, Tuple[np.ndarray, ...]]]:
        if _variables is None:
            return self.logpdf(x)
        else:
            if 'x' in _variables:
                loglike, g = self.logpdf_gradient(x)
                g = np.expand_dims(g, list(range(loglike.ndim)))
                return loglike, (g,)
            else:
                return self.logpdf(x), ()

    def as_factor(self, variable: "Variable", name: Optional[str] = None
                  ) -> "FactorJacobian":
        from autofit.graphical import FactorJacobian
        if name is None:
            shape = self.shape
            clsname = type(self).__name__
            family = clsname[:-7] if clsname.endswith('Message') else clsname
            name = f"{family}Likelihood" + (str(shape) if shape else '')

        return FactorJacobian(self, x=variable, name=name, vectorised=True)

    @classmethod
    def transformed(
            cls,
            transform: AbstractDensityTransform,
            clsname: Optional[str] = None,
            support: Optional[Tuple[Tuple[float, float], ...]] = None,
    ) -> Type["AbstractMessage"]:
        """
        transforms the distribution according the passed transform, 
        returns a newly created class that encodes the transformation.

        Parameters
        ----------
        transform: AbstractDensityTransform
            object that transforms the density
        clsname: str, optional
            the class name of the newly created class.
            defaults to "Transformed<OriginalClassName>"
        support: Tuple[Tuple[float, float], optional
            the support of the new class. Generally this can be 
            automatically calculated from the parent class

        Examples
        --------
        >>> from autofit.graphical.messages import NormalMessage, transform

        Normal distributions have infinite univariate support
        >>> NormalMessage._support
        ((-inf, inf),)

        We can tranform the NormalMessage to the unit interval 
        using `transform.phi_transform`
        >>> UnitNormal = NormalMessage.transformed(transform.phi_transform)
        >>> message = UnitNormal(1.2, 0.8)
        >>> message._support
        ((0.0, 1.0),)

        Samples from the UnitNormal will exist in the Unit interval
        >>> samples = message.sample(1000)
        >>> samples.min(), samples.mean(), samples.max()
        (0.06631750944045942, 0.8183189295040845, 0.9999056316923468)

        Projections still work for the transformed class
        >>> UnitNormal.project(samples, samples*0) 
        TransformedNormalMessage(mu=1.20273342, sigma=0.80929032)

        Can specify the name of the new transformed class
        >>> NormalMessage.transformed(transform.phi_transform, 'UnitNormal')(0, 1.)
        UnitNormal(mu=0, sigma=1.)

        The transformed objects are pickleable
        >>> import pickle
        >>> pickle.loads(pickle.dumps(message))
        TransformedNormalMessage(mu=1.2, sigma=0.8)

        The transformed objects also are normalised,
        >>> from scipy.integrate import quad
        >>> quad(message.pdf, 0, 1)
        (1.0000000000114622, 3.977073226302252e-09)

        Can also nest transforms
        >>> WeirdNormal = NormalMessage.transformed(
            transform.log_transform).transformed(
            transform.exp_transform)
        This transformation is equivalent to the identity transform!
        >>> WeirdNormal.project(NormalMessage(0.3, 0.8).sample(1000))
        Transformed2NormalMessage(mu=0.31663248, sigma=0.79426984)

        This functionality is more useful for applying linear shifts
        e.g.
        >>> ShiftedUnitNormal = NormalMessage.transformed(
            transform.phi_transform
        ).shifted(shift=0.7, scale=2.3)
        >>> ShiftedUnitNormal._support
        ((0.7, 3.0),)
        >>> samples = ShiftedUnitNormal(0.2, 0.8).sample(1000)
        >>> samples.min(), samples.mean(), samples.max()
        """
        support = support or tuple(zip(*map(
            transform.inv_transform, map(np.array, zip(*cls._support))
        ))) if cls._support else cls._support
        projectionClass = (
            None if cls._projection_class is None
            else cls._projection_class.transformed(transform)
        )
        if issubclass(cls, TransformedMessage):
            depth = cls._depth + 1
            clsname = clsname or f"Transformed{depth}{cls._Message.__name__}"

            # Don't doubly inherit if transforming already transformed message
            class Transformed(cls):  # type: ignore
                __qualname__ = clsname
                _Message = cls
                _transform = transform
                _support = support
                __projection_class = projectionClass
                _depth = depth

        else:
            clsname = clsname or f"Transformed{cls.__name__}"

            class Transformed(TransformedMessage, cls):  # type: ignore
                __qualname__ = clsname
                _Message = cls
                _transform = transform
                _support = support
                __projection_class = projectionClass
                parameter_names = cls.parameter_names
                _depth = 1

        Transformed.__name__ = clsname
        return Transformed

    @classmethod
    def shifted(cls, shift: float = 0, scale: float = 1) -> Type["AbstractMessage"]:
        return cls.transformed(
            LinearShiftTransform(shift=shift, scale=scale),
            clsname=f"Shifted{cls.__name__}"
        )

    @classmethod
    def _reconstruct(
            cls,
            parameters: Tuple[np.ndarray, ...],
            log_norm: float,
            *args
    ):
        return cls(*parameters, log_norm=log_norm)

    def __reduce__(self):
        # serialises TransformedMessage during pickling
        return (
            self._reconstruct,
            (
                self.parameters,
                self.log_norm
            ),
        )

    @classmethod
    def _sample(cls, self, n_samples):
        # Needed for nested TransformedMessage method resolution
        return cls.sample(self, n_samples)

    @classmethod
    def _logpdf_gradient(cls, self, x):
        # Needed for nested TransformedMessage method resolution
        return cls.logpdf_gradient(self, x)

    def as_prior(self):
        raise NotImplemented()


class TransformedMessage(AbstractMessage):
    _Message: Type[AbstractMessage]
    _transform: AbstractDensityTransform
    _depth = 0

    @classmethod
    def _reconstruct(  # type: ignore
            cls,
            Message: 'AbstractMessage',
            clsname: str,
            transform: AbstractDensityTransform,
            parameters: Tuple[np.ndarray, ...],
            log_norm: float
    ):
        # Reconstructs TransformedMessage during unpickling
        Transformed = Message.transformed(transform, clsname)
        return Transformed(*parameters, log_norm=log_norm)

    def __reduce__(self):
        # serialises TransformedMessage during pickling
        return (
            TransformedMessage._reconstruct,
            (
                self._Message,
                self.__class__.__name__,
                self._transform,
                self.parameters,
                self.log_norm
            ),
        )

    @classmethod
    def calc_log_base_measure(cls, x) -> np.ndarray:
        x, log_det = cls._transform.transform_det(x)
        log_base = cls._Message.calc_log_base_measure(x)
        return log_base + log_det

    @classmethod
    def to_canonical_form(cls, x) -> np.ndarray:
        x = cls._transform.transform(x)
        return cls._Message.to_canonical_form(x)

    @cached_property
    def mean(self) -> np.ndarray:
        return self._transform.inv_transform(self._Message.mean.func(self))

    @cached_property
    def variance(self) -> np.ndarray:
        return self._transform.inv_transform(self._Message.variance.func(self))

    @classmethod
    def _sample(cls, self, n_samples) -> np.ndarray:
        x = cls._Message._sample(self, n_samples)
        return cls._transform.inv_transform(x)

    @classmethod
    def _logpdf_gradient(  # type: ignore
            cls,
            self,
            x: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        x, logd, logd_grad, jac = cls._transform.transform_det_jac(x)
        logl, grad = cls._Message._logpdf_gradient(self, x)
        return logl + logd, grad * jac + logd_grad

    def sample(self, n_samples=None) -> np.ndarray:
        return self._sample(self, n_samples)

    def logpdf_gradient(
            self, x: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        return self._logpdf_gradient(self, x)

    # TODO add code for analytic hessians when Jacobian is fixed e.g. for shifted messages
    logpdf_gradient_hessian = AbstractMessage.numerical_logpdf_gradient_hessian

    @classmethod
    def from_mode(
            cls,
            mode: np.ndarray,
            covariance: np.ndarray,
            id_=None
    ) -> "AbstractMessage":
        mode, jac = cls._transform.transform_jac(mode)
        covariance = jac.invquad(covariance)
        return cls.from_mode(mode, covariance, id_=id_)
